﻿using System;
using System.Drawing;
using System.IO;
using System.Drawing.Imaging;
using System.Collections.Generic;

namespace 随机验证码
{
  public class GifCore
  {
    private MemoryStream _stream;
    private string _filePath;
    private int _delay;
    private int _repeat;

    /// <summary>
    /// 创建一个新的动画GIF
    /// </summary>
    /// <param name="delay">帧之间的延迟(毫秒)</param>
    /// <param name="repeat">GIF重复计数(0表示永久)</param>
    /// <returns></returns>
    public GifCore(int delay = 500, int repeat = 0)
    {
      this._delay = delay;
      this._repeat = repeat;
    }

    /// <summary>
    /// 为GIF添加一个新帧
    /// </summary>
    /// <param name="image">要添加到GIF堆栈中的图像</param>
    /// <param name="delay">帧之间的延迟(毫秒)(-1:使用全局设置)</param>
    public void AddFrame(Image image, int delay = -1)
    {
      GifInfo gif = new GifInfo();
      gif.LoadGifPicture(image);

      if (this._stream == null)//创建一张新的GIF89a图片
      {
        this._stream = new MemoryStream();
        Write(this._stream, CreateHeaderBlock());
        Write(this._stream, gif.LogicalScreenDescriptor.ToArray());
        Write(this._stream, CreateApplicationExtensionBlock(this._repeat));
      }

      Write(this._stream, CreateGraphicsControlExtensionBlock(delay == -1 ? this._delay : delay));
      Write(this._stream, gif.ImageDescriptor.ToArray());
      Write(this._stream, gif.ColorTable.ToArray());
      Write(this._stream, gif.ImageData.ToArray());

    }

    /// <summary>
    /// 完成创建GIF并开始刷新
    /// </summary>
    public Bitmap Finish()
    {
      if (this._stream == null)
        return null;
      this._stream.WriteByte(0x3B); // Image terminator
      Bitmap result = new Bitmap(this._stream);
      this._stream.Dispose();
      return result;
    }

    /// <summary>
    /// 创建GIF89a头部块（6个字节）
    /// </summary>
    private static byte[] CreateHeaderBlock()
    {
      //标识符 Signature (3字节） ---GIF
      //版本   Version  （3字节） ---87a (or 89a)
      return new[] { (byte)'G', (byte)'I', (byte)'F', (byte)'8', (byte)'9', (byte)'a' };
    }

    /// <summary>
    /// 创建应用程序扩充块
    /// </summary>
    /// <param name="repeat"></param>
    /// <returns></returns>
    private static byte[] CreateApplicationExtensionBlock(int repeat)
    {
      byte[] buffer = new byte[19];
      buffer[0] = 0x21; // Extension introducer
      buffer[1] = 0xFF; // Application extension
      buffer[2] = 0x0B; // Size of block
      buffer[3] = (byte)'N'; // NETSCAPE2.0
      buffer[4] = (byte)'E';
      buffer[5] = (byte)'T';
      buffer[6] = (byte)'S';
      buffer[7] = (byte)'C';
      buffer[8] = (byte)'A';
      buffer[9] = (byte)'P';
      buffer[10] = (byte)'E';
      buffer[11] = (byte)'2';
      buffer[12] = (byte)'.';
      buffer[13] = (byte)'0';
      buffer[14] = 0x03; // Size of block
      buffer[15] = 0x01; // Loop indicator
      buffer[16] = (byte)(repeat % 0x100); // 指定重复次数
      buffer[17] = (byte)(repeat / 0x100); // 0代表无穷循环
      buffer[18] = 0x00; // Block terminator
      return buffer;
    }

    /// <summary>
    /// 创建图像控制扩展块(这个块可选)（8个字节）
    /// </summary>
    /// <param name="delay"></param>
    /// <returns></returns>
    private static byte[] CreateGraphicsControlExtensionBlock(int delay)
    {
      byte[] buffer = new byte[8];
      buffer[0] = 0x21; // Extension Introducer - 标识一个extension块的开始，这个字段值为0x21。
      buffer[1] = 0xF9; // Graphic Control Label - 标识当前块为一个Graphic Control Extension。这个字段值为0xF9。
      buffer[2] = 0x04; // Block Size - 在块中的字节数，在这个Block Size Field之后，直到但是并不包括Block Terminator，这个field包含固定的值4。
      buffer[3] = 0x09; // Flags: reserved, disposal method, user input, transparent color
      buffer[4] = (byte)(delay / 10 % 0x100); // Delay time low byte
      buffer[5] = (byte)(delay / 10 / 0x100); // Delay time high byte
      buffer[6] = 0xFF; // Transparent color index
      buffer[7] = 0x00; // Block terminator
      return buffer;
    }

    private static void Write(MemoryStream stream, byte[] array)
    {
      stream.Write(array, 0, array.Length);
    }

  }

  public class GifInfo
  {
    /// <summary>
    /// 图片颜色表
    /// </summary>
    public List<byte> ColorTable = new List<byte>();
    /// <summary>
    /// 图片数据
    /// </summary>
    public List<byte> ImageData = new List<byte>();
    /// <summary>
    /// 图像描述块信息
    /// </summary>
    public List<byte> ImageDescriptor = new List<byte>();
    /// <summary>
    /// 逻辑屏幕描述符块信息
    /// </summary>
    public List<byte> LogicalScreenDescriptor = new List<byte>();

    /// <summary>
    /// 加载指定图片
    /// </summary>
    /// <param name="img"></param>
    public void LoadGifPicture(Image img)
    {
      List<byte> dataList;

      using (MemoryStream ms = new MemoryStream())
      {
        img.Save(ms, ImageFormat.Gif);
        dataList = new List<byte>(ms.ToArray());
      }

      this.AnalyzeHeader(dataList);
      this.AnalyzeLogicalScreenDescriptor(dataList);

      var blockType = this.GetTypeOfNextBlock(dataList);
      while (blockType != GifBlockType.Trailer)
      {
        switch (blockType)
        {
          case GifBlockType.ImageDescriptor:
            this.AnalyzeImageDescriptor(dataList);
            break;
          case GifBlockType.Extension:
            this.ThrowAwayExtensionBlock(dataList);
            break;
        }
        blockType = this.GetTypeOfNextBlock(dataList);
      }
    }

    /// <summary>
    /// 分析图片文件头块(6个字节)
    /// </summary>
    /// <param name="gifData"></param>
    /// <returns></returns>
    private void AnalyzeHeader(List<byte> gifData)
    {
      gifData.RemoveRange(0, 6);
    }

    /// <summary>
    /// 分析逻辑屏幕描述符块(7个字节)
    /// </summary>
    /// <param name="gifData"></param>
    private void AnalyzeLogicalScreenDescriptor(List<byte> gifData)
    {
      for (int i = 0; i < 7; i++)//提取逻辑屏幕描述块
      {
        this.LogicalScreenDescriptor.Add(gifData[i]);
      }
      gifData.RemoveRange(0, 7);

      bool globalColorTableFollows = (this.LogicalScreenDescriptor[4] & 0x80) != 0;//全局颜色表

      if (globalColorTableFollows)
      {
        int pixel = this.LogicalScreenDescriptor[4] & 0x07;
        int lengthOfColorTableInByte = 3 * (int)Math.Pow(2, pixel + 1);
        for (int i = 0; i < lengthOfColorTableInByte; i++)
        {
          this.ColorTable.Add(gifData[i]);
        }
        gifData.RemoveRange(0, lengthOfColorTableInByte);
      }

      this.LogicalScreenDescriptor[4] = (byte)(this.LogicalScreenDescriptor[4] & 0x7F);
    }

    /// <summary>
    /// 分析图像描述块(10个字节)
    /// </summary>
    /// <param name="gifData"></param>
    private void AnalyzeImageDescriptor(List<byte> gifData)
    {
      for (int i = 0; i < 10; i++)
      {
        this.ImageDescriptor.Add(gifData[i]);
      }
      gifData.RemoveRange(0, 10);

      bool localColorMapFollows = (this.ImageDescriptor[9] & 0x80) != 0;//本地颜色表
      if (localColorMapFollows)
      {
        this.ColorTable.Clear();
        int pixel = this.ImageDescriptor[9] & 0x07;
        int lengthOfColorTableInByte = 3 * (int)Math.Pow(2, pixel + 1);
        for (int i = 0; i < lengthOfColorTableInByte; i++)
        {
          this.ColorTable.Add(gifData[i]);
        }
        gifData.RemoveRange(0, lengthOfColorTableInByte);
      }
      else
      {
        int lastThreeBitsOfGlobalTableDescription = this.LogicalScreenDescriptor[4] & 0x07;
        this.ImageDescriptor[9] = (byte)(this.ImageDescriptor[9] & 0xF8);
        this.ImageDescriptor[9] = (byte)(this.ImageDescriptor[9] | lastThreeBitsOfGlobalTableDescription);
      }

      this.ImageDescriptor[9] = (byte)(this.ImageDescriptor[9] | 0x80);
      this.GetImageData(gifData);
    }

    /// <summary>
    /// 获取图片数据
    /// </summary>
    /// <param name="gifData"></param>
    private void GetImageData(List<byte> gifData)
    {
      this.ImageData.Add(gifData[0]);
      gifData.RemoveAt(0);

      while (gifData[0] != 0x00)//0x00表示后面已经没有图片数据
      {
        int countOfFollowingDataBytes = gifData[0];
        for (int i = 0; i <= countOfFollowingDataBytes; i++)
        {
          this.ImageData.Add(gifData[i]);
        }
        gifData.RemoveRange(0, countOfFollowingDataBytes + 1);
      }

      this.ImageData.Add(gifData[0]);
      gifData.RemoveAt(0);
    }

    /// <summary>
    /// 
    /// </summary>
    /// <param name="gifData"></param>
    private void ThrowAwayExtensionBlock(List<byte> gifData)
    {
      gifData.RemoveRange(0, 2); // Delete ExtensionBlockIndicator and ExtensionDetermination

      while (gifData[0] != 0)
      {
        gifData.RemoveRange(0, gifData[0] + 1);
      }
      gifData.RemoveAt(0);
    }

    /// <summary>
    /// 获取下一个区块类型
    /// </summary>
    /// <param name="gifData"></param>
    /// <returns></returns>
    private GifBlockType GetTypeOfNextBlock(List<byte> gifData)
    {
      return (GifBlockType)gifData[0];
    }

    /// <summary>
    /// 类型转换
    /// </summary>
    /// <param name="b"></param>
    /// <returns></returns>
    private char ByteToChar(byte b)
    {
      return (char)b;
    }

    /// <summary>
    /// 区块类型
    /// </summary>
    public enum GifBlockType
    {
      /// <summary>
      /// 图像描述块
      /// </summary>
      ImageDescriptor = 0x2C,
      /// <summary>
      /// 
      /// </summary>
      Extension = 0x21,
      /// <summary>
      /// 文件尾端标志块(固定为3B)
      /// </summary>
      Trailer = 0x3B
    }

  }
}